#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from ply import lex
from .lexertoken import LexerToken
from .lexer import Lexer


class LexBasedLexer(Lexer):
    def __init__(self):
        self._lexer = lex.lex(object=self)
        self._lexer.current_token = ''
        self._lexer.current_token_pos = -1

    def token(self):
        token = self._lexer.token()
        if token is None and self._lexer.lexstate != 'INITIAL':
            return LexerToken('LITERAL',
                              value=self._lexer.current_token,
                              partial=True,
                              lexpos=self._lexer.current_token_pos)
        return token

    def input(self, text):
        while self._lexer.lexstatestack:
            self._lexer.pop_state()
        return self._lexer.input(text)

    def get_position(self):
        return self._lexer.lexpos


class TemplateLexerError(Exception):
    pass


class TemplateLexer(LexBasedLexer):
    # pylint: disable=invalid-name
    tokens = (
        'LITERAL', 'MACRO', "TEMPLATE_FUNC"
    )

    states = (
        ('dollar', 'exclusive'),
        ('dollarbrace', 'exclusive'),
        ('dollarparen', 'exclusive'),
        ('string', 'exclusive'),
        ('qstring', 'exclusive'),
    )

    # Tokens

    def t_LITERAL(self, t):
        r'[^\$]+'
        t.type = "LITERAL"
        return t

    def t_DOLLAR(self, t):
        r'\$'
        t.lexer.push_state('dollar')
        t.lexer.current_token = '$'
        t.lexer.current_token_pos = t.lexpos

    def t_dollar_MACRO(self, t):
        r'[a-zA-Z0-9_]+'
        t.lexer.pop_state()
        t.value = '$' + t.value
        t.lexpos = t.lexer.current_token_pos
        return t

    def t_dollar_BRACE_OPEN(self, t):
        r'{'
        t.lexer.push_state('dollarbrace')
        t.lexer.current_token = '$' + t.value

    def t_dollarbrace_MACRO(self, t):
        r'[^}]+'
        t.lexer.current_token += t.value

    def t_dollarbrace_BRACE_CLOSE(self, t):
        r'}'
        t.lexer.current_token += t.value
        # go back to INITIAL
        t.lexer.pop_state()
        t.lexer.pop_state()
        t.type = 'MACRO'
        t.value = t.lexer.current_token
        t.lexpos = t.lexer.current_token_pos
        return t

    def t_dollar_PAREN_OPEN(self, t):
        r'\('
        t.lexer.push_state('dollarparen')
        t.lexer.paren_count = 1
        t.lexer.current_token = '$('
        t.lexer.current_token_pos = t.lexpos - 1

    def t_dollarparen_PAREN_OPEN(self, t):
        r'\('
        t.lexer.paren_count += 1
        t.lexer.current_token += '('

    def t_dollarparen_TEMPLATE_FUNC(self, t):
        r'''[^()'"]+'''
        t.lexer.current_token += t.value

    def t_dollarparen_PAREN_CLOSE(self, t):
        r'\)'
        t.lexer.current_token += ')'
        t.lexer.paren_count -= 1
        if t.lexer.paren_count == 0:
            t.type = 'TEMPLATE_FUNC'
            t.value = t.lexer.current_token
            t.lexpos = t.lexer.current_token_pos
            t.lexer.pop_state()
            t.lexer.pop_state()
            return t

        return None

    def t_dollarparen_QUOTE(self, t):
        r'"'
        t.lexer.current_token += t.value
        t.lexer.push_state('string')

    def t_dollarparen_APOSTROPHE(self, t):
        r"'"
        t.lexer.current_token += t.value
        t.lexer.push_state('qstring')

    def t_string_CHARS(self, t):
        r'[^"\\]'
        t.lexer.current_token += t.value

    def t_string_backslash_plus_character(self, t):
        r'\\.'
        t.lexer.current_token += t.value

    def t_string_QUOTE(self, t):
        r'"'

        # closing quote character
        t.lexer.current_token += t.value
        t.lexer.pop_state()

    def t_qstring_CHARS(self, t):
        r"[^']"
        t.lexer.current_token += t.value

    def t_qstring_APOSTROPHE(self, t):
        r"'"

        # closing apostrophe
        t.lexer.current_token += t.value
        t.lexer.pop_state()

    def t_error(self, t):
        raise TemplateLexerError(f"Illegal character {t.value} in state {self._lexer.lexstate}")

    def t_dollar_error(self, t):
        return self.t_error(t)

    def t_dollarbrace_error(self, t):
        return self.t_error(t)

    def t_dollarparen_error(self, t):
        return self.t_error(t)

    def t_string_error(self, t):
        return self.t_error(t)

    def t_qstring_error(self, t):
        return self.t_error(t)



#############################################################################
# Copyright (c) 2022 Balazs Scheidler
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

#
# This is a sample Python destination module that show-cases how to
# implement Python modules and Python based destinations, see
# modules/python-modules/README.md for more details
#

from syslogng import LogDestination
import logging


class ExampleDestination(LogDestination):

    logger = logging.getLogger(__name__)

    def __init__(self):
        self._is_opened = False

    def init(self, options):
        self.logger.info("ExampleDestination: initialized with {}".format(options))
        return True

    def is_opened(self):
        return self._is_opened

    def open(self):
        self.logger.info("ExampleDestination: opened")
        self._is_opened = True
        return True

    def close(self):
        self.logger.info("ExampleDestination: closed")
        self._is_opened = False

    def deinit(self):
        self.logger.info("ExampleDestination: deinit")

    def send(self, msg):
        self.logger.info("ExampleDestination: Name Value Pairs are:")

        for key in msg.keys():
            value = msg[key]
            self.logger.info("ExampleDestination: {} = {} {}".format(key, value, type(value)))
        return True



#############################################################################
# Copyright (c) 2024-2025 Axoflow
# Copyright (c) 2024 László Várady
# Copyright (c) 2025 One Identity LLC.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from syslogng import LogSource, LogMessage
from collections import defaultdict

import logging
import asyncio
import threading
import tornado.web
import ssl
import signal
import json
from typing import Any

WEBHOOK_NV_PREFIX = "webhook."
WEBHOOK_QUERY_NV_PREFIX = WEBHOOK_NV_PREFIX + "query."
WEBHOOK_HEADERS_KEY = WEBHOOK_NV_PREFIX + "headers"


class Handler(tornado.web.RequestHandler):
    def initialize(self, source) -> None:
        self.source = source

    @tornado.web.authenticated
    async def post(self, **path_arguments) -> None:
        # racy, but the client should retry
        if self.source.suspended.is_set():
            self.set_status(503)
            await self.finish({"status": "flow-controlled"})
            return

        msg = self._construct_msg(self.request, path_arguments)
        self.source.post_message(msg)

        await self.finish({"status": "received"})

    def _set_proxied_ip(self, msg: LogMessage) -> None:
        proxy_headers = self.request.headers.get_list(self.source.proxy_header)

        if proxy_headers and len(proxy_headers) > 0:
            # the closest/last IP (the proxy_header flag implies that the last one can be trusted)
            msg.set_source_ipaddress(proxy_headers[-1])
            msg["PEERIP"] = self.request.remote_ip
            return

        msg.set_source_ipaddress(self.request.remote_ip)

    def _set_request_headers(self, msg: LogMessage) -> None:
        headers = defaultdict(list)
        for h in self.request.headers.get_all():
            name = h[0]
            if name:
                headers[name].append(h[1])

        msg[WEBHOOK_HEADERS_KEY] = json.dumps(headers)

    def _construct_msg(self, request, path_arguments) -> LogMessage:
        msg = LogMessage(self.request.body)
        msg.set_recvd_rawmsg_size(len(self.request.body))

        for key, value in request.query_arguments.items():
            value = value[0] if len(value) == 1 else value
            msg[WEBHOOK_QUERY_NV_PREFIX + key] = value

        for key, value in path_arguments.items():
            msg[key] = value

        if self.source.include_request_headers:
            self._set_request_headers(msg)

        if self.source.proxy_header:
            self._set_proxied_ip(msg)
        else:
            msg.set_source_ipaddress(self.request.remote_ip)

        return msg

    def set_default_headers(self) -> None:
        self.set_header("Server", "syslog-ng")

    def get_current_user(self):
        if not self.source.auth_token:
            return True

        token = self.request.headers.get("Authorization", "").split(" ")
        if len(token) != 2:
            self.source.logger.debug("Auth failed, missing Authorization header or auth-scheme")
            return False

        token = token[1]
        return self.source.auth_token == token

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        self.set_status(status_code)


class HTTPSource(LogSource):
    def init(self, options: dict[str, Any]) -> bool:
        self.logger = logging.getLogger("http")
        self.set_transport_name("http")
        if not self.init_options(options):
            return False

        if not self.port:
            self.port = 443 if self.tls_key_file else 80
        self.port = int(self.port)

        self.ssl_ctx = None
        if self.tls_key_file:
            self.setup_tls()

        self.suspended = threading.Event()
        self.event_loop = asyncio.new_event_loop()
        self.request_exit = None

        async def init_request_exit():
            self.request_exit = asyncio.Event()
        self.event_loop.run_until_complete(init_request_exit())

        handlers = list(map(lambda p: (p, Handler, {"source": self}), self.paths))
        if len(handlers) == 0:
            handlers = [(r"/.*", Handler, {"source": self})]

        self.app = tornado.web.Application(
            handlers,
            log_function=self.log_access,
            autoreload=False,
            compress_response=False,
        )
        self.server = None

        return True

    async def runServer(self) -> None:
        self.logger.info(f"HTTP(S) server started on port {self.port}")
        self.server = self.app.listen(self.port, decompress_request=True, ssl_options=self.ssl_ctx)
        await self.request_exit.wait()

        self.server.stop()
        await self.server.close_all_connections()

    async def stopServer(self) -> None:
        self.request_exit.set()

    def deinit(self) -> None:
        pass

    def run(self) -> None:
        self.event_loop.run_until_complete(self.runServer())

    def suspend(self) -> None:
        self.suspended.set()

    def wakeup(self) -> None:
        self.suspended.clear()

    def request_exit(self) -> None:
        asyncio.run_coroutine_threadsafe(self.stopServer(), self.event_loop)
        pass

    def log_access(self, req: tornado.web.RequestHandler) -> None:
        self.logger.debug(f"{req.get_status()} {req._request_summary()}")

    def setup_tls(self) -> None:
        self.ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.ssl_ctx.load_cert_chain(certfile=self.tls_cert_file, keyfile=self.tls_key_file)

        if self.tls_peer_verify:
            self.logger.debug("Enabling client cert verification")
            self.ssl_ctx.verify_mode = ssl.CERT_REQUIRED
        else:
            self.ssl_ctx.verify_mode = ssl.CERT_NONE

        if self.tls_use_system_cert_store:
            self.logger.debug("Using system cert store for client auth")
            self.ssl_ctx.load_default_certs(ssl.Purpose.CLIENT_AUTH)

        self.tls_ca_dir = self.tls_ca_dir if self.tls_ca_dir else None
        self.tls_ca_file = self.tls_ca_file if self.tls_ca_file else None

        if self.tls_ca_dir or self.tls_ca_file:
            self.ssl_ctx.load_verify_locations(cafile=self.tls_ca_file, capath=self.tls_ca_dir)

    def init_options(self, options: dict[str, Any]) -> bool:
        try:
            self.port = options.get("port")
            self.auth_token = options.get("auth_token")
            self.paths = options.get("paths", [])
            if isinstance(self.paths, str):
                self.paths = [self.paths]

            self.tls_key_file = options.get("tls_key_file")
            self.tls_cert_file = options.get("tls_cert_file")

            self.tls_peer_verify = bool(options.get("tls_peer_verify", False))
            self.tls_use_system_cert_store = bool(options.get("tls_use_system_cert_store", False))
            self.tls_ca_file = options.get("tls_ca_file")
            self.tls_ca_dir = options.get("tls_ca_dir")

            self.proxy_header = options.get("proxy_header")
            if self.proxy_header == "yes":
                self.proxy_header = "x-forwarded-for"

            self.include_request_headers = bool(options.get("include_request_headers", False))

            return True
        except KeyError as e:
            self.logger.error(f"Missing option '{e.args[0]}'")
            return False



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .completer import Completer


class LangBasedCompleter(Completer):

    def __init__(self, parser, completers, prefix=None):
        self._parser = parser
        self._completers = completers
        self._prefix = prefix or ''
        self._completing_at_the_beginning_of_the_grammar = False

    def complete(self, entire_text, word_to_be_completed):
        if entire_text.startswith(self._prefix):
            return self._handle_input_with_prefix(entire_text, word_to_be_completed)
        return self._handle_input_without_prefix(entire_text)

    def _handle_input_without_prefix(self, entire_input):
        if self._prefix.startswith(entire_input) or entire_input == '':
            return [self._prefix]
        return []

    def _handle_input_with_prefix(self, entire_input, word_to_be_completed):
        entire_input, word_to_be_completed = self._chop_prefixes(entire_input, word_to_be_completed)

        expected_tokens, last_token, completion_prefix, word_to_be_completed = (
            self._evaluate_language(entire_input, word_to_be_completed))

        completers = self._find_applicable_completers(expected_tokens)
        completions = self._collect_completions(completers, last_token, completion_prefix, word_to_be_completed)
        return completions

    def _chop_prefixes(self, entire_input, word_to_be_completed):
        if word_to_be_completed == entire_input:
            word_to_be_completed = word_to_be_completed[len(self._prefix):]
            self._completing_at_the_beginning_of_the_grammar = True
        else:
            self._completing_at_the_beginning_of_the_grammar = False
        entire_input = entire_input[len(self._prefix):]
        return entire_input, word_to_be_completed

    def _evaluate_language(self, entire_input, word_to_be_completed):
        expected_tokens, replaced_token, token_position = (
            self._parser.get_expected_tokens(entire_input, drop_last_token=word_to_be_completed != ''))

        if token_position >= 0:
            word_position = len(entire_input) - len(word_to_be_completed)
            if word_position <= token_position:
                completion_prefix = entire_input[word_position:token_position]
                word_to_be_completed = entire_input[token_position:]
            else:
                completion_prefix = ''
        else:
            completion_prefix = ''

        if replaced_token is None:
            replaced_token = ''
        else:
            replaced_token = replaced_token.value or ''

        return expected_tokens, replaced_token, completion_prefix, word_to_be_completed

    def _find_applicable_completers(self, expected_tokens):
        completers = []
        for token in expected_tokens:
            try:
                completers.append(self._completers[token])
            except KeyError:
                pass
        return completers

    def _collect_completions(self, completers, last_token, completion_prefix, word_to_be_completed):
        if self._completing_at_the_beginning_of_the_grammar:
            prefix = self._prefix + completion_prefix
        else:
            prefix = completion_prefix
        completions = []
        for completer in completers:
            completions.extend(
                (prefix + completion for completion in completer.complete(last_token, word_to_be_completed)))
        completions = sorted(completions)
        return completions



#############################################################################
# Copyright (c) 2025 One Identity LLC
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
import json
from os import path, unlink
from queue import SimpleQueue, Full, Empty
from uuid import uuid4

from .compressable_file_buffer import CompressableFileBuffer

from pathlib import Path
from threading import Lock
from typing import Any, Dict, Callable
from logging import getLogger


class S3ObjectBuffer:

    logger = getLogger("S3.ObjectBuffer.S3ObjectBuffer")

    def __init__(self, working_dir: Path, object_key: str, target_index: int, target_timestamp: str, creator: str, object_settings: Dict[str, Any]) -> None:
        self.__buffer_lock = Lock()
        self.__meta = {
            "creator": creator,
            "object_key": object_key,
            "suffix": object_settings.get("suffix", ""),
            "target_index": target_index,
            "target_timestamp": target_timestamp,
            "storage_class": object_settings.get("storage_class", "STANDARD"),
            "compression": object_settings.get("compression", False),
            "compresslevel": object_settings.get("compresslevel", 0),
            "max_size": object_settings.get("max_object_size", None),
            "acl": object_settings.get("canned_acl", None),
            "content_type": object_settings.get("content_type", "application/octet-stream"),
        }
        self.__basename = f"{object_key}_{uuid4()}"
        self.__path = Path(working_dir, self.__basename)
        self.__meta_path = Path(working_dir, self.__basename + "_meta.json")
        self.retries = 0
        self.buffer = None
        self.__buffer_source = "pending"
        self.__finished = False

    def create_file(self):
        self.buffer = CompressableFileBuffer(self.__path, self.__meta["compression"], self.__meta["compresslevel"])
        with open(self.__meta_path, "w") as meta_file:
            meta_file.write(json.dumps(self.__meta))
        self.__buffer_source = "created"

    def load_metadata(self, meta_path: Path) -> None:
        self.__meta_path = meta_path
        self.__path = Path(str(meta_path).removesuffix("_meta.json"))
        with open(self.__meta_path, "r") as meta_file:
            self.__meta = self.__meta | json.loads(meta_file.read())
        self.buffer = CompressableFileBuffer(self.__path, self.__meta["compression"], self.__meta["compresslevel"])
        self.__buffer_source = "loaded"

    def finish(self):
        with self.__buffer_lock:
            self.__finish()

    def __finish(self):
        if not self.__finished:
            self.buffer.finish()
            self.buffer.close()
            self.__finished = True

    def write(self, msg: bytes) -> bool:
        if self.__finished:
            return False
        with self.__buffer_lock:
            if self.buffer is None:
                self.create_file()
            if not self.buffer.write(msg):
                return False
            if self.buffer.tell() >= self.__meta["max_size"]:
                self.__finish()
        return True

    @property
    def creator(self) -> str:
        return self.__meta["creator"]

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def object_key(self) -> str:
        return self.__meta["object_key"]

    @property
    def full_target_key(self) -> str:
        if self.__meta["target_index"] == 0:
            full_key = self.__meta["object_key"] + self.__meta["suffix"]
        else:
            full_key = f"{self.__meta['object_key']}-{self.__meta['target_index']}{self.__meta['suffix']}"
        return full_key

    @property
    def target_timestamp(self) -> str:
        return self.__meta["target_timestamp"]

    @property
    def last_modified(self):
        return int(path.getmtime(self.__path))

    @property
    def buffer_created(self):
        return self.__buffer_source

    @property
    def finished(self):
        if self.buffer is None:
            return False
        else:
            return self.__finished

    def deprecate(self):
        with self.__buffer_lock:
            self.buffer.unlink()
            unlink(self.__meta_path)

    @property
    def target_index(self):
        return self.__meta["target_index"]

    @property
    def storage_class(self):
        return self.__meta["storage_class"]

    @property
    def acl(self):
        return self.__meta["acl"]

    @property
    def content_type(self):
        return self.__meta["content_type"]


class S3ObjectQueue:

    logger = getLogger("S3.ObjectBuffer.S3ObjectQueue")

    def __init__(self):
        self.queue: SimpleQueue[S3ObjectBuffer] = SimpleQueue()

    def enqueue(self, s3_object: S3ObjectBuffer, enqueue_cb: Callable[[], None] | None):
        try:
            self.queue.put(s3_object)
        except Full:
            self.logger.error("Object queue is full, could not enqueue.")
            return
        if enqueue_cb:
            enqueue_cb()

    def dequeue(self) -> S3ObjectBuffer | None:
        try:
            return self.queue.get_nowait()
        except Empty:
            return None



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .lexertoken import LexerToken
from .lexer import Lexer

CLL_NORMAL = 0
CLL_QUOTED_STRING = 1
CLL_QUOTED_CHAR = 2


class CommandLineLexer(Lexer):
    # pylint: disable=assignment-from-no-return
    """This an inperfect lexer for both the debug language and template functions"""
    def __init__(self):
        self._tokens = []
        self._input = ''
        self._current_token = 0
        self._current_state = CLL_NORMAL
        self._paren_balance = 0
        self._current_position = 0
        self._quote_open_character = ''

    def input(self, text):
        self._input = text
        self._current_position = 0
        self._current_state = CLL_NORMAL
        self._paren_balance = 0

    def token(self):
        token = self._get_next_token()
        return token

    def get_position(self):
        return self._current_position

    def _get_next_token(self):
        self._skip_whitespace()
        self._current_token = ''
        self._paren_balance = 0
        start_position = self._current_position
        token = None

        while self._current_position < len(self._input):
            current_char = self._input[self._current_position]

            if self._current_state == CLL_NORMAL:
                token = self._process_normal_character(current_char)
            elif self._current_state == CLL_QUOTED_STRING:
                token = self._process_string_character(current_char)
            elif self._current_state == CLL_QUOTED_CHAR:
                token = self._process_escaped_character(current_char)

            self._current_position += 1
            if token is not None:
                return token

        if start_position != self._current_position:
            partial = False
            if self._current_state != CLL_NORMAL:
                partial = True
            if self._paren_balance != 0:
                partial = True
            return LexerToken('ARG', value=self._current_token, partial=partial, lexpos=start_position)

        return None

    def _skip_whitespace(self):
        while self._current_position < len(self._input) and self._input[self._current_position].isspace():
            self._current_position += 1

    def _process_normal_character(self, current_char):
        if current_char in ('"', "'"):
            return self._open_quoted_string(current_char)
        elif current_char.isspace():
            return self._close_current_token(current_char)
        elif current_char == '(':
            return self._open_paren(current_char)
        elif current_char == ')':
            return self._close_paren(current_char)
        else:
            self._current_token += current_char

        return None

    def _open_paren(self, current_char):
        self._paren_balance += 1
        self._current_token += current_char

    def _close_paren(self, current_char):
        self._paren_balance -= 1
        if self._paren_balance <= 0:
            self._current_token += current_char
            return LexerToken('ARG', value=self._current_token)
        self._current_token += current_char

        return None

    def _close_current_token(self, current_char):
        if self._paren_balance == 0:
            return LexerToken('ARG', value=self._current_token)
        self._current_token += current_char

        return None

    def _process_string_character(self, current_char):
        if current_char == '\\':
            self._current_state = CLL_QUOTED_CHAR
        elif current_char == self._quote_open_character:
            self._close_quoted_string(current_char)
        else:
            self._current_token += current_char

    def _process_escaped_character(self, current_char):
        self._current_token += current_char
        self._current_state = CLL_QUOTED_STRING

    def _open_quoted_string(self, current_char):
        self._quote_open_character = current_char
        self._current_state = CLL_QUOTED_STRING
        if self._paren_balance > 0:
            self._current_token += current_char

    def _close_quoted_string(self, current_char):
        if self._paren_balance > 0:
            self._current_token += current_char
        self._current_state = CLL_NORMAL



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import absolute_import, print_function
from .langcompleter import LangBasedCompleter
from .debuglang import DebugLang
from .templatelang import TemplateLang
from .tflang import TemplateFunctionLang
from .choicecompleter import ChoiceCompleter
from .macrocompleter import MacroCompleter
from .syslognginternals import get_debugger_commands, get_template_functions, get_nv_registry, get_value_pairs_scopes


class DebuggerCLI():
    debug_completers = {}
    template_completers = {}
    tflang_completers = {}

    def _setup_tflang_completers(self):
        self.tflang_completers['COMMAND'] = ChoiceCompleter(get_template_functions())
        self.tflang_completers['name_value_name'] = ChoiceCompleter(get_nv_registry())
        self.tflang_completers['value_pairs_scope'] = ChoiceCompleter(get_value_pairs_scopes())
        self.tflang_completers['OPT'] = ChoiceCompleter(TemplateFunctionLang.known_options)
        self.tflang_completers['template'] = LangBasedCompleter(parser=TemplateLang(),
                                                                completers=self.template_completers)

    def _setup_template_completers(self):
        self.template_completers['MACRO'] = MacroCompleter()
        self.template_completers['TEMPLATE_FUNC'] = LangBasedCompleter(
            parser=TemplateFunctionLang(),
            completers=self.tflang_completers,
            prefix="$(")

    def _setup_debug_completers(self):
        self.debug_completers['COMMAND'] = ChoiceCompleter(get_debugger_commands())
        self.debug_completers['template'] = LangBasedCompleter(parser=TemplateLang(),
                                                               completers=self.template_completers)

    def get_root_completer(self):
        self._setup_debug_completers()
        self._setup_template_completers()
        self._setup_tflang_completers()
        return LangBasedCompleter(parser=DebugLang(),
                                  completers=self.debug_completers)
